# ===================
#
#       process
#
# ===================

# -------------------
# ---- command reader
# -------------------
proc getcommand {} \
{
  if {[eof $::channel]} { _exit } \
  else \
  { 
    # check time-out
    watchdog
    # register next enregs
    fileevent $::channel readable putbuffer
    # read command
    set cmd [gets $::channel]
    # execute command
    switch -exact -- [lindex $cmd 0] \
    {
      ""      { }
      :admin: { eval $cmd }
      default { execute $cmd }
    }
  }
} 
# -------------------
# ---- command executer
# -------------------
proc execute {cmd} \
{
  log >>$cmd
  set key [lindex $cmd 0]
  set args [lrange $cmd 1 end]
  if {[info exists ::intern($key)] && $::intern($key) != ""} \
  { # internal commands
    # ------------
    set rc [script { set res [eval $::intern($key) $args] }]
  } \
  elseif {[info exists ::extern($key)] && $::extern($key) != ""} \
  { # external commands
    # ------------
    set rc [script { set res [eval exec $::extern($key) $args] }]
  } \
  else \
  { # unknown command
    # ------------
      set res [doService $cmd]
    #  set rc [script { set res [eval $cmd] }]
    # logerr "***unknown command"
  }
  if {[info exists res]} { output $res }
  output :end:
  # wait for next command
  fileevent $::channel readable getcommand
}

proc doService {script} {
    set out  ""
    set message "" 
    set result [catch {
	uplevel #0 $script
    } message]
    if ($result==0) {
	set out $message   
    } else {
	set out "error: $message"   
    }
    puts $out
    return $out
}

# -------------------
# ---- buffer writer
# -------------------
proc putbuffer {} \
{
  if {[eof $::channel]} { _exit } \
  else \
  {
    # get enreg
    set text [gets $::channel]
    switch -exact -- [lindex $text 0] \
    {
      :admin: \
      { # server request
        eval $text 
      }
      default \
      { # data from client
        set id [incr ::buffer(PutID)]
        # save for later
        set ::buffer($id) $text
      }
    }
    watchdog
  }
} 
# -------------------
# ---- buffer reader
# -------------------
proc getbuffer {} \
{
  set putid $::buffer(PutID)
  set getid [incr ::buffer(GetID)]
  if {$putid < $getid} { vwait ::buffer($getid) }
  # return saved data
  return $::buffer($getid)
}
# -------------------
# ---- channel writer
# -------------------
proc output {text} \
{ 
  set text [string map {\n \\n} $text]
  set rc [script { puts $::channel $text; flush $::channel }]
  if {$rc} { catch { close $::channel } }
  return $rc
}
# -------------------
# ---- logger
# -------------------
set ::log 0
proc log {text} \
{ 
  if {$::log} { return }
  set ::log 1
  # send log to server
  if {[regexp {^>>admin} $text]} \
  { set text [lreplace $text 2 2 ********] }
  output [list :log: "[_date]#$::client $text"] 
  set ::log 0
}

proc logerr {text} \
{
  # log an error
  catch \
  { 
    set text [list :error: $text]
    output $text 
  }
}
# -------------------
# ---- error displayer
# -------------------
proc bgerror {msg} { logerr $msg }
# -------------------
# ---- watchdog
# -------------------
proc watchdog {} \
{
  # check for time-out
  set delay $::parms(watchdog)
  if {$delay != 0} \
  {
    after cancel _exit 1
    after $delay _exit 1
  }
}
# -------------------
# ---- internal commands
# -------------------
  # ---- help
proc _help {args} \
{
  output "internal commands:\t[join [lsort [array names ::intern]] \t]"
  output "external commands:\t[join [lsort [array names ::extern]] \t]"
  return ""
}

proc _ping {} \
{
    return "pong"
}

proc _log {} \
{
    set fp [open "out-$::ip-$::client.log" r]
    set data [read $fp]
    close $fp
    return $data
}

# ---- admin
proc _admin {user passwd} \
{
  # -- admin login
  if {$user != $::parms(user) || $passwd != $::parms(passwd)} \
  { 
    output [list :error: "bad login"]
    return ""
  }
  # -- admin subcommands
  while 1 \
  {
    set cmd [getbuffer]
    log --$cmd
    switch -exact -- [lindex $cmd 0] \
    {
      quit    { break }
      :admin: { eval $cmd }
      default { output ":admin: $cmd" }
    }
    output :end:
  }
  return :end:
}
proc :admin: {cmd args} \
{
  # server request
  log ":admin: $cmd $args"
  switch -exact -- $cmd \
  {
    :init:    { catch { source $::parms(init) } }
    default   { output [list :error: "unknown admin request \"$cmd\""] }
  }
}
# -------------------
# ---- admin commands
# -------------------
array set admin \
{
  parms     _parms
  intern    _intern
  extern    _extern
  help      _help
  init      _init
  count     _count
  clients   _clients
}
  # ---- stop
set ::exited 0
proc _exit {{timeout 0}} \
{
  if {!$::exited} \
  { 
    set ::exited 1
    if {$timeout} \
    { output [list :error: "process timed out after $::parms(watchdog)"] }
    log "exiting process"
  }
  catch { close $::channel }
  exit
}
# -------------------
# ---- dater
# -------------------
set ::date(Last) 0
proc _date {} \
{
  set time [clock seconds]
  if {$time != $::date(Last)} \
  {
    set ::date(Last) $time
    set ::date(Format) [clock format $time -format {%Y-%m-%d %H:%M:%S}]
  }
  return $::date(Format)
}

# -------------------
# ---- script executer
# -------------------
proc script {script} \
{
  # catch & display error, if any
  set rc [catch { uplevel 1 $script } msg]
  if {$rc} { logerr $msg } 
  return $rc
}
# -------------------
# ---- error logger
# -------------------
proc bgerror {args} { logerr $args }
# -------------------
# ---- set parameters
# -------------------
  # default values
array set intern \
{
  admin     _admin
  help      _help
  date      _date
  ping      _ping
  log       _log
  exit      _exit
}
array set extern \
{
  df        df
  du        du
  ps        ps
  ls        ls
}
array set parms \
{
  port      3456
  admin     3457
  watchdog  604800000
  echo      1
  log       server.log
  init      .server.tcl
  user      server
  passwd    tcl
  tclsh     /usr/local/svi/tcl/bin/tclsh8.5
  process   process.tcl
  font      {Courier 9}
}
# -------------------
# ---- start process
# -------------------
set arg0 [string range [lindex $argv 0] 1 end]
set arg1 [string range [lindex $argv 1] 1 end]
set ::ip [string range [lindex $argv 2] 1 end]

set ::client $arg1
# get local parms
script { source $arg0 }
# init buffer
set ::buffer(PutID) 0
set ::buffer(GetID) 0
# init channel
if {![script { set ::channel [socket localhost $parms(admin)] }]} \
{
  # init channel
  watchdog
  fconfigure $::channel -blocking 0
  fileevent  $::channel readable getcommand
  # send command
  output ":process: $arg1"
  # initial log
  log "starting process"
  # enter event loop
  vwait forever
}
