# ===================
#
#       server
#
# ===================

# -------------------
# ---- client connection
# -------------------
proc connect {channel client port} \
{ 
  # log event
  log "connect $client $channel" 1
  # start a new process
  puts "channel $channel"
#  exec /usr/local/svi/tcl/bin/tclsh8.5 $::parms(process) $::parms(init) $channel &
  exec /usr/local/svi/bin/huscript -noTk $::parms(process) -$::parms(init) -$channel -$client > "out-$client-$channel.log" ?> "err.log" &
  set ::clients($channel) [list [_date] $client]
}
# -------------------
# ---- client disconnection
# -------------------
proc disconnect {client process} \
{
  catch { close $client }
  catch { close $process }
  set logfile "out-[lindex $::clients($client) 1]-$client.log" 
  if {[info exists ::clients($client)]} \
  {
    log "disconnect [lindex $::clients($client) 1] $client" 1
    unset ::clients($client)
  }
  catch {file delete $logfile}
}
# -------------------
# ---- admin channel listener
# -------------------
proc admin {channel client port} \
{
  # configure admin channel
  fconfigure $channel -blocking 0
  fileevent  $channel readable [list input $channel]
}
# -------------------
# ---- admin channel reader
# -------------------
proc input {process} \
{ 
  if {[eof $process]} { catch { close $process } } \
  else \
  {
    # get admin command
    set cmd [gets $process]
    if {[lindex $cmd 0] != ":process:"} \
    { log "***bad admin input: $cmd"; return }
    # tie a process to its client
    set client [lindex $cmd 1]
      # from process
    proc _$process {process client} \
    { 
      if {[eof $process]} \
      { 
        ##display loc "process eof $process $client"
        # closed socket
        disconnect $client $process
        rename _$process ""
      } \
      else \
      { 
        set text [gets $process]
        ##display loc "process \"$text\""
        switch -exact -- [lindex $text 0] \
        {
          :quit:  \
          { # end of process
            disconnect $client $process
          }
          :log:   \
          { # text to log
            adminlog [join [lrange $text 1 end]]
          }
          :admin: \
          { # admin request
            set key   [lindex $text 1]
            set args  [lrange $text 2 end]
            if {![info exists ::admin($key)]} \
            { 
              set text [list :error: "unknown admin request \"$key\""]
              display_err $client $text
            } \
            elseif {![script { set res [eval $::admin($key) $args] }]} \
            { output $client $res }
          }
          :error: \
          { # display, log & send error
            display_err $client $text
          }
          default \
          { # send to client
            if {$::parms(echo) && $text != ":end:"} \
            { display ok [string map {\\n \n} $text] }
            output $client $text 
          }
        }
      }
    }
    fileevent $process readable [list _$process $process $client]
      # from client
    proc _$client {client process} \
    {
      if {[eof $client]} \
      {
        ##display loc "client eof $process $client"
        # closed socket
        disconnect $client $process
        rename _$client ""
      } \
      else \
      {
        set text [gets $client]
        ##display loc "client \"$text\""
        switch -exact -- [lindex $text 0] \
        {
          :admin: \
          { # security filter
            output $client "***unknown command"
          }
          default \
          { # client request
            if {$::parms(echo)} \
            { 
              if {[lindex $text 0] == "admin"} \
              { display ok "[lrange $text 0 1] ********" } \
              else { display ok $text }
            }
            output $process $text
          }
        }
      }
    }
    fileevent $client readable [list _$client $client $process]
    # register process
    lappend ::clients($client) $process
  }
} 
# -------------------
# ---- channel writer
# -------------------
proc output {channel text} \
{ # write to channel
  set rc [script { puts $channel $text; flush $channel }]
  if {$rc} { catch { close $channel } }
  return $rc
}
# -------------------
# ---- logger
# -------------------
proc display_err {client text} \
{
  catch { display err [string map {\\n \n} [lindex $text 1]] }
  adminlog  "[_date]#$client $text"
  output $client $text 
}
proc log {text {echo 0}} \
{ 
  # echo event upon the console
  if {$echo || $::parms(echo)} { display sys $text }
  # log event inside a file
  adminlog "[_date] $text"
}
set ::log 0
proc adminlog {text} \
{ # writing the log file
  if {![info exists ::logfp] || $::log} { return }
  set ::log 1
  script \
  { 
    puts $::logfp $text
    flush $::logfp 
  } 
  set ::log 0
}
# -------------------
# ---- admin commands
# -------------------
array set admin \
{
  parms     _parms
  intern    _intern
  extern    _extern
  help      _help
  init      _init
  count     _count
  clients   _clients
}
  # ---- stop
set ::exited 0
proc _exit {} \
{ 
  if {!$::exited} \
  { set ::exited 1; log "exiting server"; exit }
}

  # ---- parms
interp alias {} _parms  {} _set parms
  # ---- intern
interp alias {} _intern {} _set intern
  # ---- extern
interp alias {} _extern {} _set extern
  # ---- set
proc _set {array args} \
{
  if {[llength $args] == 0} \
  { # display all
    set keys [array names ::$array] 
  } \
  else \
  { # get or set value
    set keys  [lindex $args 0]
    set value [lindex $args 1] 
    if {[llength $args] == 2} { set ::${array}($keys) $value }
  }
  # display result
  set res {}
  foreach key $keys \
  { 
    if {$key == "passwd"} { lappend res "passwd\t********" } \
    else { lappend res "$key\t[set ::${array}($key)]" }
  }
  return [join $res \n]
}
  # ---- help
proc _help {} \
{ return [join [lsort [concat [array names ::admin] help quit]] \t] }
  # ---- init
proc _init {} \
{
  # save local values
  set fn $::parms(init)
  set fp [open $fn w]
  puts $fp "# $fn modified at [_date]"
  puts $fp ""
  foreach array {::parms ::intern ::extern} \
  {
    puts $fp "\narray set $array \\"
    puts $fp "\{"
    foreach key [lsort [array names $array]] \
    {
      puts $fp "[format {%-20s} $key]\"[set ${array}($key)]\""
    }
    puts $fp "\}"
  }
  close $fp
  # reinit running processes
  foreach client [array names ::clients] \
  { output [lindex $::clients($client) 2] [list :admin: :init:] }
  return ""
}
  # ---- count
proc _count {} \
{ return [llength [array names ::clients]] }
  # ---- clients
proc _clients {} \
{
  set res {}
  foreach client [lsort [array names ::clients]] \
  {
    set enreg $::clients($client)
    lappend res "[lindex $enreg 0]\t[lindex $enreg 1]\t$client\t[lindex $enreg 2]"
  }
  return [join $res \n]
}
# -------------------
# ---- dater
# -------------------
set date(Last) 0
proc _date {} \
{
  set time [clock seconds]
  if {$time != $::date(Last)} \
  {
    set ::date(Last) $time
    set ::date(Format) [clock format $time -format {%Y-%m-%d %H:%M:%S}]
  }
  return $::date(Format)
}
# -------------------
# ---- script executer
# -------------------
proc script {script} \
{ # catch & display error (if any)
  set rc [catch { uplevel 1 $script } msg]
  if {$rc} { display err $msg; log $msg } 
  return $rc
}
# -------------------
# ---- error displayer
# -------------------
proc bgerror {args} { display err $args; log $args }
# -------------------
# ---- text displayer
# -------------------
set tk [expr {![catch {package require Tk}]}]
if {$tk} \
{
  proc display {{color ok} {text ""}} \
  { 
    if {$color == "sys"} { set text "[_date] $text" }
    .f.t insert end $text\n $color
    .f.t see end
  }
} \
else \
{
  proc display {{color ok} {text ""}} \
  { 
    switch -exact -- $color \
    {
      ok    { puts $text }
      sys   { puts stderr "[_date] $text" }
      err   { puts stderr "error\n$text" }
    }
  }
}
# -------------------
# ---- set parameters
# -------------------
  # default values
array set intern \
{
  admin     _admin
  help      _help
  date      _date
  ping      _ping
  exit      _exit
}
array set extern \
{
  ps        ps
  ls        ls
  cd        cd   
}
array set parms \
{
  port      3456
  admin     3457
  watchdog  604800000
  echo      1
  log       server.log
  init      .server.tcl
  user      server
  passwd    tcl
  tclsh     /usr/local/svi/tcl/bin/tclsh8.5
  process   process.tcl
  font      {Courier 9}
}
  # local values
set fn [lindex $argv 0]
if {$fn != "" && [string index $fn 0] != "-"} \
{ # first arg was init file name
  set argv [lrange $argv 1 end]
} \
else { set fn $parms(init) }
script { source $fn }
# current parms values
set keys [array names parms]
foreach {key value} $argv \
{
  set key2 [string range $key 1 end]
  if {[lsearch $keys $key] != -1} \
  { set $parms($key2) $value; display ok "-->$key2 $value" } \
  else { display err "unknown key: $key $value\n(expected: $keys)" }
}
# -------------------
# ---- init log
# -------------------
proc openlog {} \
{
  script \
  { 
    # open log file
    set ::logfp [open $::parms(log) a]
    # close & reopen each minute
    after 60000 reopenlog
  }
}
proc reopenlog {} \
{
  catch { close $::logfp; unset ::logfp }
  openlog
}
openlog
# -------------------
# ---- init display
# -------------------
if {$tk} \
{
  # create widgets
  wm title . server
  pack [frame .f]
  pack [scrollbar .f.s -command {.f.t yview}] -fill y -side right
  pack [text .f.t -yscrollcommand {.f.s set} -font $::parms(font) -wrap word]
  .f.t tag configure sys -foreground brown
  .f.t tag configure ok  -foreground navy
  .f.t tag configure err -foreground red
  pack [button .b -text exit -width 10 -command _exit2] -pady 10
  proc _exit2 {} \
  {
    set rc [tk_dialog .exit Exit "Do you really want to quit?" question 1 yes no]
    if {!$rc} { _exit }
  }
  set config 0
  proc _config {} \
  {
    # resize when needed
    if {$::config} { return }
    set ::config 1
    set rw [winfo width .]
    set rh [winfo height .]
    set rs [wm state .] 
    if {![info exists ::dw]} \
    {
      # first time
      update
      set tw [winfo width .f.t]
      set ::cw [.f.t cget -width]
      set ::dw [expr {int($tw / $::cw)}]
      set ::dx [expr {$rw - $::dw * $::cw}]
      set th [winfo height .f.t]
      set ::ch [.f.t cget -height]
      set ::dh [expr {int($th / $::ch)}]
      set ::dy [expr {$rh - $::dh * $::ch}]
    } \
    elseif {$::rs == "zoomed" && $::rs != $rs} \
    {
      .f.t config -width $::cw -height $::ch
      update
    } \
    else \
    {
      # other times
      set cw [expr {int(($rw - $::dx) / $::dw)}]
      set ch [expr {int(($rh - $::dy) / $::dh)}]
      .f.t config -width $cw -height $ch
      if {$rs != "zoomed"} { set ::cw $cw; set ::ch $ch }
    }
    set ::rs $rs 
    set ::config 0
  }
  bind . <Destroy> _exit
  bind . <Configure> _config
  wm protocol . WM_DELETE_WINDOW _exit2
}
# -------------------
# ---- start server
# -------------------
set rc [script \
{ 
  # open clients & processes sockets
  socket -server connect $parms(port) 
  socket -server admin $parms(admin) 
}]
if {!$rc} \
{
  # initial display
  log "server ready" 1
  display
  # enter event loop
  vwait forever
}

